/* arg.hh
 *
 * Copyright 2004,2005 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Revision: 2005-03-18
 *  - add SetValue
 */

#ifndef ARG_HH
#define ARG_HH 1

#include <vector>
#include <map>
#include <string>
#include <sstream>

namespace arg {

	class Value
	{
	public:
		virtual ~Value();
		virtual void set(std::string str) = 0;
	};

	typedef bool (CallBack)(int, std::string, void *);

	class Option
	{
	public:
		Option(int key, std::string name);
		~Option();

		Option & store(Value * ptr); // store value to "* ptr"
		Option & optional(std::string str = ""); // value is optional defaulting to "str"
		Option & set(int * var, int value); // set "* var" to "value"
		Option & once(int init = 0); // can only be set once, with distinct value, "init"
		Option & call(CallBack * func, void * data); // call function "* func" with "data" as extra argument
		Option & help(std::string text, std::string var = ""); // help text

		bool take_value();
		bool need_value();
		int get_key();
		std::string get_name();

		enum HelpFormat {
			HF_REGULAR,
			HF_NODASH
		};
		std::string get_help(HelpFormat format = HF_REGULAR);

		void process();
		void process(std::string str);
	private:
		int key;
		std::string name;

		Value * store_ptr; // pointer to storage space
		bool store_optional; // if value string is optional
		std::string store_str; // default value string

		int * set_var; // variable to set
		int set_value; // value to set
		bool set_once;  // if can only set once
		int set_init;  // initial value, 

		CallBack * call_func; // callback function
		void * call_data; // data to pass to callback function

		std::string help_text;
		std::string help_var;
	};

	class Parser
	{
	public:
		~Parser();
		void add_help(std::string msg);
		Option & add_opt(int key, std::string name, bool hide = false);
		std::vector<std::string> & args();
		bool parse(int argc, char * argv[]);

		std::string get_help();

		// default options
		Option & add_opt_help();
		Option & add_opt_version(std::string version);
	protected:
		std::vector<Option *> opt_list;
		std::vector<std::string> arg_list;
		struct HelpLine {
			std::string msg;
			Option * opt;
		};
		std::vector<HelpLine> help_list;
	private:
		std::string version_info;
	};

	class SubParser :
		public Value,
		public Parser
	{
	public:
		Option & add_opt(std::string name);
		virtual void set(std::string str);

		std::string get_help();

		// default options
		Option & add_opt_help();
	};

	// Errors:

	class Error
	{
	public:
		Error();
		Error(std::string msg);
		std::string get_msg();
	protected:
		std::string msg;
	};

	class OptError : // option processing error
		public Error
	{
	public:
		OptError(std::string opt);
		OptError(std::string opt, std::string msg);
	protected:
		std::string opt;
	};

	class ConvError : // conversion error
		public Error
	{
	public:
		ConvError(std::string str, std::string type);
	};

	class UnknError : // unknow option error
		public Error
	{
	public:
		UnknError(std::string opt);
	};

	// Templates:

	template <typename T> class StreamableValue :
		public Value
	{
	public:
		StreamableValue(T * t)
		{
			ptr = t;
		}
		virtual ~StreamableValue()
		{
		}
		virtual void set(std::string str)
		{
			std::istringstream s(str);
			T tmp;
			s >> tmp;
			if (s.bad() || ! s.eof()) throw ConvError(str, typeid(T).name());
			* ptr = tmp;
		}
	private:
		T * ptr;
	};

	template <typename T> StreamableValue<T> * SV(T * t)
	{
		return new StreamableValue<T>(t);
	}

	// Extensions:

	// Set of names in strings
	class SetValue :
		public Value
	{
	public:
		SetValue(int * var);
		void add(std::string name, int value);

		virtual void set(std::string str);
	private:
		int * var;
		std::map<std::string, int> set_list;
	};
}
#endif
